import XCTest
@testable import WooCommerce
@testable import Yosemite

/// Tests for `BulkUpdatePriceSettingsViewModel`
///
final class BulkUpdatePriceSettingsViewModelTests: XCTestCase {

    private var storesManager: MockStoresManager!

    override func setUp() {
        super.setUp()
        storesManager = MockStoresManager(sessionManager: SessionManager.makeForTesting())
    }

    override func tearDown() {
        storesManager = nil
        super.tearDown()
    }

    func test_initial_viewModel_state() throws {
        // Given
        let viewModel = BulkUpdatePriceSettingsViewModel(siteID: 0,
                                                         productID: 0,
                                                         bulkUpdateOptionsModel: BulkUpdateOptionsModel(productVariations: []),
                                                         editingPriceType: .regular,
                                                         priceUpdateDidFinish: { })

        // Then
        XCTAssertEqual(viewModel.saveButtonState, .disabled)
        XCTAssertNil(viewModel.bulkUpdatePriceError)
        XCTAssertEqual(viewModel.sections.count, 1)
        XCTAssertEqual(viewModel.sections.first?.rows.count, 1)
        let footerText = try XCTUnwrap(viewModel.sections.first?.footer)
        XCTAssertTrue(footerText.isNotEmpty)
    }

    func test_state_when_price_is_changed_from_empty_to_a_value() {
        // Given
        let viewModel = BulkUpdatePriceSettingsViewModel(siteID: 0,
                                                         productID: 0,
                                                         bulkUpdateOptionsModel: BulkUpdateOptionsModel(productVariations: []),
                                                         editingPriceType: .regular,
                                                         priceUpdateDidFinish: { })
        viewModel.handlePriceChange("42")

        // Then
        XCTAssertEqual(viewModel.saveButtonState, .enabled)
        XCTAssertNil(viewModel.bulkUpdatePriceError)
    }

    func test_state_when_price_is_changed_from_a_value_to_empty() {
        // Given
        let viewModel = BulkUpdatePriceSettingsViewModel(siteID: 0,
                                                         productID: 0,
                                                         bulkUpdateOptionsModel: BulkUpdateOptionsModel(productVariations: []),
                                                         editingPriceType: .regular,
                                                         priceUpdateDidFinish: { })
        viewModel.handlePriceChange("")

        // Then
        XCTAssertEqual(viewModel.saveButtonState, .disabled)
        XCTAssertNil(viewModel.bulkUpdatePriceError)
    }

    func test_state_when_no_regular_price_is_selected_and_save_button_tapped_given_variations_have_sale_price() {
        // Given
        let variations = [MockProductVariation().productVariation().copy(dateOnSaleStart: Date(), dateOnSaleEnd: Date(), salePrice: "42")]
        let viewModel = BulkUpdatePriceSettingsViewModel(siteID: 0,
                                                         productID: 0,
                                                         bulkUpdateOptionsModel: BulkUpdateOptionsModel(productVariations: variations),
                                                         editingPriceType: .regular,
                                                         priceUpdateDidFinish: { })

        // When
        viewModel.saveButtonTapped()

        // Then
        XCTAssertEqual(viewModel.saveButtonState, .disabled)
        XCTAssertEqual(viewModel.bulkUpdatePriceError?.validationError(), .salePriceWithoutRegularPrice)
    }

    func test_state_when_no_regular_price_is_selected_and_save_button_tapped_given_with_multiple_variations_having_sale_price() {
        // Given
        let variations = [MockProductVariation().productVariation().copy(dateOnSaleStart: Date(), dateOnSaleEnd: Date(), regularPrice: "43", salePrice: "42"),
                          MockProductVariation().productVariation().copy(dateOnSaleStart: Date(), dateOnSaleEnd: Date(), salePrice: "42")]
        let viewModel = BulkUpdatePriceSettingsViewModel(siteID: 0,
                                                         productID: 0,
                                                         bulkUpdateOptionsModel: BulkUpdateOptionsModel(productVariations: variations),
                                                         editingPriceType: .regular,
                                                         priceUpdateDidFinish: { })

        // When
        viewModel.saveButtonTapped()

        // Then
        XCTAssertEqual(viewModel.saveButtonState, .disabled)
        XCTAssertEqual(viewModel.bulkUpdatePriceError?.validationError(), .salePriceWithoutRegularPrice)
    }

    func test_state_when_selected_regular_price_is_less_than_sale_price() {
        // Given
        let variations = [MockProductVariation().productVariation().copy(dateOnSaleStart: Date(), dateOnSaleEnd: Date(), salePrice: "42")]
        let viewModel = BulkUpdatePriceSettingsViewModel(siteID: 0,
                                                         productID: 0,
                                                         bulkUpdateOptionsModel: BulkUpdateOptionsModel(productVariations: variations),
                                                         editingPriceType: .regular,
                                                         priceUpdateDidFinish: { })

        // When
        viewModel.handlePriceChange("10")
        viewModel.saveButtonTapped()

        // Then
        XCTAssertEqual(viewModel.saveButtonState, .enabled)
        XCTAssertEqual(viewModel.bulkUpdatePriceError?.validationError(), .salePriceHigherThanRegularPrice)
    }

    func test_state_when_selected_regular_price_is_less_than_regular_price_with_multiple_variations() {
        // Given
        let variations = [MockProductVariation().productVariation().copy(dateOnSaleStart: Date(), dateOnSaleEnd: Date(), salePrice: "1"),
                          MockProductVariation().productVariation().copy(dateOnSaleStart: Date(), dateOnSaleEnd: Date(), salePrice: "42")]
        let viewModel = BulkUpdatePriceSettingsViewModel(siteID: 0,
                                                         productID: 0,
                                                         bulkUpdateOptionsModel: BulkUpdateOptionsModel(productVariations: variations),
                                                         editingPriceType: .regular,
                                                         priceUpdateDidFinish: { })

        // When
        viewModel.handlePriceChange("10")
        viewModel.saveButtonTapped()

        // Then
        XCTAssertEqual(viewModel.saveButtonState, .enabled)
        XCTAssertEqual(viewModel.bulkUpdatePriceError?.validationError(), .salePriceHigherThanRegularPrice)
    }

    func test_state_when_selected_valid_price_is_valid_and_action_is_dispatched() {
        // Given
        let variations = [MockProductVariation().productVariation().copy(dateOnSaleStart: Date(), dateOnSaleEnd: Date(), salePrice: "9")]
        storesManager.whenReceivingAction(ofType: ProductVariationAction.self) { _ in
            // do nothing to stay in "syncing" state
        }
        let viewModel = BulkUpdatePriceSettingsViewModel(siteID: 0,
                                                         productID: 0,
                                                         bulkUpdateOptionsModel: BulkUpdateOptionsModel(productVariations: variations),
                                                         editingPriceType: .regular,
                                                         priceUpdateDidFinish: { })

        // When
        viewModel.handlePriceChange("50")
        viewModel.saveButtonTapped()

        // Then
        XCTAssertEqual(viewModel.saveButtonState, .loading)
        XCTAssertNil(viewModel.bulkUpdatePriceError)
    }

    func test_state_remains_loading_when_price_changes() {
        // Given
        let variations = [MockProductVariation().productVariation().copy(dateOnSaleStart: Date(), dateOnSaleEnd: Date(), salePrice: "9")]
        storesManager.whenReceivingAction(ofType: ProductVariationAction.self) { _ in
            // do nothing to stay in "syncing" state
        }
        let viewModel = BulkUpdatePriceSettingsViewModel(siteID: 0,
                                                         productID: 0,
                                                         bulkUpdateOptionsModel: BulkUpdateOptionsModel(productVariations: variations),
                                                         editingPriceType: .regular,
                                                         priceUpdateDidFinish: { })

        // When
        viewModel.handlePriceChange("50")
        viewModel.saveButtonTapped()
        viewModel.handlePriceChange("500")

        // Then
        XCTAssertEqual(viewModel.saveButtonState, .loading)
        XCTAssertNil(viewModel.bulkUpdatePriceError)
    }

    func test_state_when_selected_valid_price_is_valid_when_action_fails() {
        // Given
        let variations = [MockProductVariation().productVariation().copy(dateOnSaleStart: Date(), dateOnSaleEnd: Date(), salePrice: "9")]
        storesManager.whenReceivingAction(ofType: ProductVariationAction.self) { action in
            switch action {
            case let .updateProductVariations(_, _, _, onCompletion):
                onCompletion(.failure(.unexpected))
            default:
                XCTFail("Unsupported Action")
            }
        }
        let viewModel = BulkUpdatePriceSettingsViewModel(siteID: 0,
                                                         productID: 0,
                                                         bulkUpdateOptionsModel: BulkUpdateOptionsModel(productVariations: variations),
                                                         editingPriceType: .regular,
                                                         priceUpdateDidFinish: { },
                                                         storesManager: storesManager)

        // When
        viewModel.handlePriceChange("42")
        viewModel.saveButtonTapped()

        // Then
        XCTAssertEqual(viewModel.saveButtonState, .enabled)
        XCTAssertEqual(viewModel.bulkUpdatePriceError, .priceUpdateError)
    }

    func test_callback_is_called_when_update_action_is_successful() {
        // Given
        let variations = [MockProductVariation().productVariation().copy(dateOnSaleStart: Date(), dateOnSaleEnd: Date(), salePrice: "9")]
        var isCallbackCalled = false
        storesManager.whenReceivingAction(ofType: ProductVariationAction.self) { action  in
            switch action {
            case let .updateProductVariations(_, _, _, onCompletion):
                onCompletion(.success([]))
            default:
                XCTFail("Unsupported Action")
            }
        }
        let viewModel = BulkUpdatePriceSettingsViewModel(siteID: 0,
                                                         productID: 0,
                                                         bulkUpdateOptionsModel: BulkUpdateOptionsModel(productVariations: variations),
                                                         editingPriceType: .regular,
                                                         priceUpdateDidFinish: {
                                                            isCallbackCalled = true
                                                         },
                                                         storesManager: storesManager)

        // When
        viewModel.handlePriceChange("42")
        viewModel.saveButtonTapped()


        // Then
        waitUntil {
            isCallbackCalled
        }
        XCTAssertTrue(isCallbackCalled)
    }

    func test_action_is_caled_with_the_updated_price() {
        // Given
        let dateOnSaleStart = Date()
        let dateOnSaleEnd = dateOnSaleStart
        let variations = [MockProductVariation().productVariation().copy(dateOnSaleStart: dateOnSaleStart, dateOnSaleEnd: dateOnSaleEnd, salePrice: "9")]
        var updatedVariations: [ProductVariation] = []
        storesManager.whenReceivingAction(ofType: ProductVariationAction.self) { action  in
            switch action {
            case let .updateProductVariations(_, _, variations, onCompletion):
                updatedVariations = variations
                onCompletion(.success([]))
            default:
                XCTFail("Unsupported Action")
            }
        }
        let viewModel = BulkUpdatePriceSettingsViewModel(siteID: 0,
                                                         productID: 0,
                                                         bulkUpdateOptionsModel: BulkUpdateOptionsModel(productVariations: variations),
                                                         editingPriceType: .regular,
                                                         priceUpdateDidFinish: { },
                                                         storesManager: storesManager)

        // When
        viewModel.handlePriceChange("42")
        viewModel.saveButtonTapped()


        // Then
        waitUntil {
            updatedVariations.isNotEmpty
        }
        let expectedVariation = variations.map { $0.copy(regularPrice: "42") }
        assertEqual(updatedVariations, expectedVariation)
    }
}

private extension BulkUpdatePriceSettingsViewModel.BulkUpdatePriceError {
    /// Convenient method to access the associated error value
    ///
    func validationError() -> ProductPriceSettingsError? {
        switch self {
        case let .inputValidationError(validationError):
            return validationError
        default:
            return nil
        }
    }
}
